;* -------------------------------------------------------------------------
;* Imaginary Processors -- Using the AT89C2051 as a Virtual Machine
;* Appendix B. Software Listing
;* Listing 1. Local console functions coded in Virtual Machine
;* assembly language.
;* -------------------------------------------------------------------------
;* Virtual Machine C-FLEA library support for the Mid-Tech I2C 
;* user console device
;* J. C. Dybowski - 1995
;* C-FLEA instruction set and mnemonics copyright by Dave Dunfield 1994-1995

;*--------------------------------------------------------------------------*
;* Constants
;*--------------------------------------------------------------------------*
;* LCD port IIC address
?ladd   EQU     $42

;* Keypad port IIC address
?kadd   EQU     $46

;* Beeper port IIC address
?badd   EQU     $46

;* Defined I/O bits
?den    EQU     %00100000
?drs    EQU     %00010000
;*--------------------------------------------------------------------------*
;* Write a string to the lc display: lcputstr(char *string)
;*--------------------------------------------------------------------------*
lcputstr        
	LDI     2,S             ;Get argument
?lcputstr
	LDB     I               ;Get character
	SJZ     ?7              ;Exit if null
	PUSHI                   ;Save pointer
	PUSHA                   ;Pass parm
	CALL    ?lcputch        ;Output it
	LDB     S+              ;Clean stack
	LDI     S+              ;Restore pointer
	LEAI    1,I             ;Advance
	SJMP    ?lcputstr       ;Proceed until done
;*--------------------------------------------------------------------------*
;* Write a character to the lc display: lcputch(char c)
;*--------------------------------------------------------------------------*
lcputch
	LDB     2,S             ;Get character
?lcputch
	CMPB    #$A             ;Newline?
	SJNZ    ?10             ;Yes
;* Adjust cursor first if necessary
	LDB     ?cursor
	LEAI    ?1              ;Point to table
	SWITCH                  ;Dispatch
?1      FDB     ?2,20           ;End of first line
	FDB     ?3,40           ;End of second line
	FDB     ?4,60           ;End of third line
	FDB     ?5,80           ;End of fourth line
	FDB     0,?6            ;Default
?2
	LDB     #64+%10000000   ;Set to line 2
	CALL    ?comm_wr        ;Fix cursor
	SJMP    ?6
?3
	LDB     #20+%10000000   ;Set to line 3
	CALL    ?comm_wr        ;Fix cursor
	SJMP    ?6
?4
	LDB     #84+%10000000   ;Set to line 4
	CALL    ?comm_wr        ;Fix cursor
	SJMP    ?6              
?5
	LDB     #00+%10000000   ;Set to line 1
	CALL    ?comm_wr        ;Fix cursor
	CLR
	STB     ?cursor         ;Home cursor
;* Character output
?6                                      
	LDB     2,S             
	CALL    ?data_wr        ;Emit character
	LDB     ?cursor
	INC                     ;Track cursor
	STB     ?cursor 
?7
	RET
;* Handle newline character
?10
	LDB     ?cursor
	CMPB    #20             ;On line 1?
	UGT
	SJNZ    ?11
	LDB     #20             ;Yes, set to line 2
	SJMP    ?14
?11
	LDB     ?cursor
	CMPB    #40             ;On line 2?
	UGT
	SJNZ    ?12
	LDB     #40             ;Yes, set to line 3
	SJMP    ?14
?12
	LDB     ?cursor
	CMPB    #60             ;On line 3?
	UGT
	SJNZ    ?13
	LDB     #60             ;Yes, set to line 4
	SJMP    ?14
?13             
	CLR                     ;Set to line 1 by default
?14
	PUSHA                   ;Pass parm
	CALL    ?lcposition     ;Set to line 0
	LDB     S+              ;Clean stack
	RET
;*--------------------------------------------------------------------------*
;* Clear display: lcblank()
;*--------------------------------------------------------------------------*
lcblank
	LDB     #1
	CALL    ?comm_wr        ;Clear command
	CLR
	STB     ?cursor         ;Cursor to 0
	LDB     #1
	CALL    ?delay
	RET
;*--------------------------------------------------------------------------*
;* Position cursor: lcposition(char c)
;*--------------------------------------------------------------------------*
lcposition
	LDB     2,S             ;Get position
?lcposition
	STB     ?cursor
	CMPB    #60             ;= 60 ?
	UGE
	SJZ     ?21      
	LDB     2,S             ;Get position
	SUBB    #60
	ADD     #84+%10000000   ;Set on line 4
	CALL    ?comm_wr
	SJMP    ?24
?21     
	LDB     2,S             ;Get position
	CMPB    #40             ;= 40 ?
	UGE
	SJZ     ?22
	LDB     2,S             ;Get position
	SUBB    #40
	ADD     #20+%10000000   ;Set on line 3
	CALL    ?comm_wr
	SJMP    ?24
?22
	LDB     2,S             ;Get position
	CMPB    #20             ;= 20 ?
	UGE
	SJZ     ?23
	LDB     2,S             ;Get position
	SUBB    #20
	ADD     #64+%10000000   ;Set on line 2
	CALL    ?comm_wr
	SJMP    ?24
?23
	LDB     2,S             ;Get position
	ADD     #00+%10000000   ;Set on line 1
	CALL    ?comm_wr
?24
	RET
;*--------------------------------------------------------------------------*
;* Select cursor type: lccursor(char c)
;* Where: 0=invisible, 1=underline, 2=blinking
;*--------------------------------------------------------------------------*
lccursor
	LDB     2,S             ;Get type
	LEAI    ?31             ;Point to table
	SWITCH                  
?31     FDB     ?32,0           ;Invisible
	FDB     ?33,1           ;Underline
	FDB     ?34,2           ;Blinking
	FDB     0,?35           ;Default
?32
	LDB     #%00001100 
	CALL    ?comm_wr
	RET
?33
	LDB     #%00001110
	CALL    ?comm_wr
	RET
?34
	LDB     #%00001111
	CALL    ?comm_wr
?35
	RET
;*--------------------------------------------------------------------------*
;* lc display initialization: lcinit()
;*--------------------------------------------------------------------------*
lcinit
	LDB     #5
	CALL    ?delay          ;Initial delay
	CALL    ?41             ;Reset LSI
	CALL    ?41             ;Reset LSI
	CALL    ?41             ;Reset LSI
	LDB     #%0010          ;4 bit mode
	ORB     #?den
	CALL    ?lc_wr
	ANDB    #255-?den
	CALL    ?lc_wr
	LDB     #1
	CALL    ?delay
	LDB     #%00101100      ;4 bit, 2 line, 4x7 matrix
	CALL    ?comm_wr      
	LDB     #1
	CALL    ?delay
	LDB     #%00001100      ;Display on, cursor off
	CALL    ?comm_wr
	LDB     #1
	CALL    ?delay
	LDB     #%00000110      ;Auto-increment, shift right
	CALL    ?comm_wr
	LDB     #1
	CALL    ?delay
	CALL    lcblank         ;Initialize cursor
	RET
;* Put LCD LSI in known state
?41
	LDB     #%0011          ;Reset lsi
	ORB     #?den
	CALL    ?lc_wr
	ANDB    #255-?den
	CALL    ?lc_wr
	LDB     #1
	CALL    ?delay
	RET
;*--------------------------------------------------------------------------*
;* Check for a character from the IIC keypad: int chkch()
;*--------------------------------------------------------------------------*
kchkch
	IN      ?kadd
	PUSHA                   ;Save scan code
	ANDB    #%10000000
	SJNZ    ?51             ;No data or I2C error
	SJMP    ?kgetch
?51
	LDB     S+              ;Clean stack
	CLR                     ;Nothing to return
	RET
;*--------------------------------------------------------------------------*
;* Get a character from the IIC keypad: int kgetch()
;*--------------------------------------------------------------------------*
kgetch 
	IN      ?kadd
	PUSHA                   ;Save scan code
	ANDB    #%10000000
	SJZ     ?kgetch         ;Data is available
	LDB     S+              ;Clean stack
	SJMP    kgetch          ;No data or I2C error
?kgetch
	LDB     #%10011111
	OUT     ?kadd           ;Assert ACK strobe and enable beep
	LDB     #%11111111
	OUT     ?kadd           ;Clear ACK strobe and disable beeper
	LDB     S+              ;Retrieve scan code
	ANDB    #$F             ;Mask unused bits
	LEAI    ?61             ;Point to table
	ADAI                    ;Generate displacement
	LDB     I               ;Translate to ASCII
	RET        
;* Scan code to ASCII translation table
?61
	STR     'ABCD'
	STR     '369'
	DB      0AH
	STR     '2580'
	STR     '147*'
;*--------------------------------------------------------------------------*
;* Keypad initialization
;*--------------------------------------------------------------------------*
kinit
	LDB     #%11111111
	OUT     ?kadd           ;Clear ACK strobe
	LDB     #%10111111
	OUT     ?kadd           ;Assert ACK strobe
	LDB     #%11111111
	OUT     ?kadd           ;Clear ACK strobe
	RET
;*--------------------------------------------------------------------------*
;* Generate variable duration beep sound: beep(int i)
;*--------------------------------------------------------------------------*
beep
	LDB     #%11011111
	OUT     ?badd           ;Enable beeper
	LD      2,S             ;Get duration
?71
	PUSHA
	LD      #10
	CALL    ?delay          ;Invoke delay
	LD      S+
	DEC                     ;Decrement multiplier
	SJNZ    ?71             ;Check multiplier
	LDB     #%11111111
	OUT     ?badd           ;Disable beeper
	RET

;*****  Support functions follow *****

;* Write to lcd command register
?comm_wr
	PUSHA
	ANDB    #$F0
	SHR     #4
	ANDB    #255-?drs
	ORB     #?den
	CALL    ?lc_wr
	ANDB    #255-?den
	CALL    ?lc_wr
	LDB     S+
	ANDB    #$F
	ANDB    #255-?drs
	ORB     #?den
	CALL    ?lc_wr
	ANDB    #255-?den
	CALL    ?lc_wr
	RET

;* Write to lcd data register
?data_wr
	PUSHA
	ANDB    #$F0
	SHR     #4
	ORB     #?drs
	ORB     #?den
	CALL    ?lc_wr
	ANDB    #255-?den
	CALL    ?lc_wr
	LDB     S+
	ANDB    #$0F
	ORB     #?drs
	ORB     #?den
	CALL    ?lc_wr
	ANDB    #255-?den
	CALL    ?lc_wr
	RET

;* Output a byte over IIC to lcd
?lc_wr
	PUSHA
	OUT     ?ladd           ;Output it
	LDB     S+
	RET

;* General delay, ACC contains multiplier
?delay
	DEC
	SJNZ    ?delay  
	RET

$DD:?cursor     1               ;Cursor location storage

